//================================================================================================================================
//
// Copyright (c) 2015-2019 VisionStar Information Technology (Shanghai) Co., Ltd. All Rights Reserved.
// EasyAR is the registered trademark or trademark of VisionStar Information Technology (Shanghai) Co., Ltd in China
// and other countries for the augmented reality technology developed by VisionStar Information Technology (Shanghai) Co., Ltd.
//
//================================================================================================================================

package cn.easyar.samples.helloarrecording;

import android.Manifest;
import android.annotation.TargetApi;
import android.content.pm.PackageManager;
import android.content.res.Configuration;
import android.os.Build;
import android.os.Bundle;

import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.view.ViewGroup;
import android.view.WindowManager;
import android.util.Log;
import android.widget.Button;
import android.widget.Toast;

import java.io.File;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;

import cn.easyar.FunctorOfVoidFromPermissionStatusAndString;
import cn.easyar.FunctorOfVoidFromRecordStatusAndString;
import cn.easyar.PermissionStatus;
import cn.easyar.RecordStatus;
import cn.easyar.Engine;
import android.app.Activity;
import android.widget.Toast;


public class MainActivity extends Activity
{
    /*
    * Steps to create the key for this sample:
    *  1. login www.easyar.com
    *  2. create app with
    *      Name: HelloARRecording
    *      Package Name: cn.easyar.samples.helloarrecording
    *  3. find the created item in the list and show key
    *  4. set key string bellow
    */
    private static String key = "===PLEASE ENTER YOUR KEY HERE===";
    private GLView glView;
    private String url = "";
    private boolean started = false;

    private String prepareUrl()
    {
        String timeStamp = new SimpleDateFormat("yyyy-MM-dd_HH-mm-ss").format(new Date());
        String FileName = "EasyAR_Recording_" + timeStamp + ".mp4";
        File folder = new File("/sdcard/Movies");
        if (!folder.exists())
            folder.mkdirs();
        return "/sdcard/Movies/" + FileName;
    }

    @Override
    protected void onCreate(Bundle savedInstanceState)
    {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_main);
        getWindow().setFlags(WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON, WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON);

        if (!Engine.initialize(this, key)) {
            Log.e("HelloAR", "Initialization Failed.");
            Toast.makeText(MainActivity.this, Engine.errorMessage(), Toast.LENGTH_LONG).show();
            return;
        }

        glView = new GLView(this);

        requestCameraPermission(new PermissionCallback() {
            @Override
            public void onSuccess() {
                ((ViewGroup) findViewById(R.id.preview)).addView(glView, new ViewGroup.LayoutParams(ViewGroup.LayoutParams.MATCH_PARENT, ViewGroup.LayoutParams.MATCH_PARENT));

                Button recordStart  = (Button)findViewById(R.id.ib_start );
                Button recordStop  = (Button)findViewById(R.id.ib_stop);
                recordStart.setOnClickListener(new View.OnClickListener() {
                    @Override
                    public void onClick(View view) {
                        if (started) {
                            return;
                        }
                        glView.requestPermissions(new FunctorOfVoidFromPermissionStatusAndString() {
                            @Override
                            public void invoke(int status, String msg) {

                                switch (status)
                                {
                                    case PermissionStatus.Denied:
                                        started = false;
                                        Log.e("HelloAR", "Permission Denied" + msg);
                                        Toast.makeText(MainActivity.this, "Permission Denied", Toast.LENGTH_SHORT).show();
                                        break;
                                    case PermissionStatus.Error:
                                        started = false;
                                        Log.e("HelloAR", "Permission Error" + msg);
                                        Toast.makeText(MainActivity.this, "Permission Error", Toast.LENGTH_SHORT).show();
                                        break;
                                    case PermissionStatus.Granted:
                                        Log.i("HelloAR", "Permission Granted");

                                        url = prepareUrl();

                                        glView.startRecording(url, new FunctorOfVoidFromRecordStatusAndString() {
                                            @Override
                                            public void invoke(int status, String value) {
                                                switch (status)
                                                {
                                                    case RecordStatus.OnStarted:
                                                        break;
                                                    case RecordStatus.OnStopped:
                                                        started = false;
                                                        break;
                                                    case RecordStatus.FileFailed:
                                                        started = false;
                                                        break;
                                                    case RecordStatus.FailedToStart:
                                                        started = false;
                                                        Toast.makeText(MainActivity.this, value, Toast.LENGTH_LONG).show();
                                                        break;
                                                    default:
                                                        break;

                                                }
                                                Log.i("HelloAR", "Recorder Callback status: " + Integer.toString(status) + ", MSG: " + value);
                                            }
                                        });
                                        Toast.makeText(MainActivity.this, "Recording...", Toast.LENGTH_SHORT).show();
                                        break;
                                }

                            }
                        });
                        started = true;
                    }
                });
                recordStop.setOnClickListener(new View.OnClickListener() {
                    @Override
                    public void onClick(View view) {
                        if (!started) {
                            return;
                        }
                        started = false;
                        glView.stopRecording();
                        Toast.makeText(MainActivity.this, "Recorded at " + url, Toast.LENGTH_LONG).show();
                    }
                });
            }

            @Override
            public void onFailure() {
            }
        });
    }

    private interface PermissionCallback
    {
        void onSuccess();
        void onFailure();
    }
    private HashMap<Integer, PermissionCallback> permissionCallbacks = new HashMap<Integer, PermissionCallback>();
    private int permissionRequestCodeSerial = 0;
    @TargetApi(23)
    private void requestCameraPermission(PermissionCallback callback)
    {
        if (Build.VERSION.SDK_INT >= 23) {
            if (checkSelfPermission(Manifest.permission.CAMERA) != PackageManager.PERMISSION_GRANTED) {
                int requestCode = permissionRequestCodeSerial;
                permissionRequestCodeSerial += 1;
                permissionCallbacks.put(requestCode, callback);
                requestPermissions(new String[]{Manifest.permission.CAMERA}, requestCode);
            } else {
                callback.onSuccess();
            }
        } else {
            callback.onSuccess();
        }
    }

    @Override
    public void onRequestPermissionsResult(int requestCode, String[] permissions, int[] grantResults)
    {
        if (permissionCallbacks.containsKey(requestCode)) {
            PermissionCallback callback = permissionCallbacks.get(requestCode);
            permissionCallbacks.remove(requestCode);
            boolean executed = false;
            for (int result : grantResults) {
                if (result != PackageManager.PERMISSION_GRANTED) {
                    executed = true;
                    callback.onFailure();
                }
            }
            if (!executed) {
                callback.onSuccess();
            }
        }
        super.onRequestPermissionsResult(requestCode, permissions, grantResults);
    }

    @Override
    public void onConfigurationChanged(Configuration newConfig) {
        super.onConfigurationChanged(newConfig);

        if (!started) {
            return;
        }
        started = false;
        glView.stopRecording();
        Toast.makeText(MainActivity.this, "Recorded at " + url, Toast.LENGTH_LONG).show();
    }

    @Override
    protected void onResume()
    {
        super.onResume();
        if (glView != null) { glView.onResume(); }
    }

    @Override
    protected void onPause()
    {
        if (glView != null) { glView.onPause(); }
        super.onPause();
    }
}
